﻿<#
.SYNOPSIS
    Update the ETW manifest files in the instrumentation folder.

.DESCRIPTION
    The script will stop the Monitoring Agent, unregister the existing ETW manifest files in the
    instrumentation folder, update the ETW manifests in the instrumentation folder, and run the
    MonitoringInstall scheduled task to register the updated ETW manifests and start the Monitoring
    Agent. MonitoringInstall-Servicing.psm1 must exist in the script folder.

    Exit codes:
       0: Success.
       2: Monitoring PowerShell module could not be initialized.
       4: Monitoring Agent could not be stopped.
       8: Monitoring Agent could not be started.
      16: One or more manifest files could not be copied to the backup folder.
      32: One or more manifest files could not be updated.
    1024: Unknown error.

    Copyright © 2019 Microsoft. All rights reserved.
#>
[CmdletBinding()]
Param(
    [Parameter(Mandatory = $false, HelpMessage = "The maximum number of seconds to wait for the stop tasks to complete.")]
    [int]$StopTaskMaxWaitSec = 300,
    [Parameter(Mandatory = $false, HelpMessage = "The maximum number of seconds to wait for the start task to complete. Set to 0 to not wait.")]
    [int]$StartTaskMaxWaitSec = 300,
    [Parameter(Mandatory = $false, HelpMessage = "The path to the log directory in which output should be written.")]
    [string]$LogDir
)

# Initialize exit code.
[int]$ExitCode = 0

# Initialize Monitoring module and logging.
[string]$LogPath = $null
try
{
    if ($LogDir)
    {
        # Add log file name.
        $LogPath = Join-Path -Path $LogDir -ChildPath "Monitoring-EtwUpdate_$([DateTime]::UtcNow.ToString("yyyyMMddHHmmss")).log"

        if ([System.IO.Path]::IsPathRooted($LogPath))
        {
            # Ensure that the directory exists.
            $LogParentPath = Split-Path -Path $LogPath -Parent
            if (!(Test-Path -Path $LogParentPath -ErrorAction SilentlyContinue))
            {
                New-Item -Path $LogParentPath -ItemType "Directory" | Out-Null
            }
        }
        else
        {
            # Resolve relative path under current directory.
            $LogPath = Join-Path -Path $PSScriptRoot -ChildPath $LogPath
        }
    }
    $MonitoringModulePath = Join-Path -Path $PSScriptRoot -ChildPath "MonitoringInstall-Servicing.psm1"
    if (!(Test-Path -Path $MonitoringModulePath -ErrorAction SilentlyContinue))
    {
        throw "Monitoring servicing module not found at: $($MonitoringModulePath)"
    }

    # Import Monitoring module.
    Import-Module -Name $MonitoringModulePath

    # For Monitoring Write-Message - Set log file path.
    Set-WriteMessageLogFilePath -Path $LogPath
}
catch
{
    $ExitCode = 2
    $ErrorMessage = "Error initializing Monitoring PowerShell module from '$($MonitoringModulePath)': $($_)"

    # Use Write-Output since Write-Message is defined in module that could not be loaded.
    Write-Output "$([DateTime]::UtcNow.ToString("yyyy-MM-dd HH:mm:ss")): $($ErrorMessage)"
    Write-Output "$($_.ScriptStackTrace)"
    Write-Output "$($_.Exception)"

    # If a log path was defined, also write error message to it.
    if ($LogPath)
    {
        "$([DateTime]::UtcNow.ToString("yyyy-MM-dd HH:mm:ss")): $($ErrorMessage)" | Out-File -FilePath $LogPath -Append
        "$($_.ScriptStackTrace)" | Out-File -FilePath $LogPath -Append
        "$($_.Exception)" | Out-File -FilePath $LogPath -Append
        "ETW update script failed with exit code: $($ExitCode)." | Out-File -FilePath $LogPath -Append
    }

    # Use throw to indicate error to AXUpdateInstaller.
    # In case of exceptions, the output is not captured, so only the error message and
    # log file contents will be available for diagnostics.
    throw "$($ErrorMessage) [Log: $($LogPath)]"
}

try
{
    Write-Message "Script to update Monitoring ETW manifest files starting..."
    Write-Message "Command: $(@([Environment]::GetCommandLineArgs()) -join " ")" -Vrb

    # Get the settings file path.
    $SettingsFilePath = Join-Path -Path $PSScriptRoot -ChildPath "Servicing.settings"

    # Check if ETW manifest update is applicable from settings.
    [bool]$UpdateApplicable = $true
    if (Test-Path -Path $SettingsFilePath -ErrorAction SilentlyContinue)
    {
        Write-Message "Reading settings from: $($SettingsFilePath)"
        $SettingsJson = Get-Content -Path $SettingsFilePath -Raw | ConvertFrom-Json

        if ($SettingsJson.ETWManifestUpdate.Skip)
        {
            $UpdateApplicable = $false
            Write-Message "Settings indicate that ETW manifest update is not applicable. Skipping ETW manifest update step."
        }
        else
        {
            Write-Message "Settings indicate that ETW manifest update is applicable."
        }
    }

    # Check if ETW manifest update is applicable from files in ETWManifest folder.
    if ($UpdateApplicable)
    {
        # Script parent path.
        $ScriptParentPath = Split-Path -Parent $PSScriptRoot

        # Source path of updated manifest files.
        $ETWManifestPath = Join-Path -Path $ScriptParentPath -ChildPath "ETWManifest"
        Write-Message "ETW manifest source path: $($ETWManifestPath)"

        if (Test-Path -Path $ETWManifestPath -ErrorAction SilentlyContinue)
        {
            $ETWManifestFiles = @(Get-ChildItem -Path $ETWManifestPath -File)
            if ($ETWManifestFiles.Count -eq 0)
            {
                $UpdateApplicable = $false
                Write-Message "ETW manifest path does not contain any files. Skipping ETW manifest update step."
            }
        }
        else
        {
            $UpdateApplicable = $false
            Write-Message "ETW manifest path does not exist. Skipping ETW manifest update step."
        }
    }

    # If applicable, stop ETW session and agent, backup and update manifest files, register ETW events, and start agent.
    if ($UpdateApplicable)
    {
        # Stop the Monitoring Agent ETW sessions and agent processes.
        try
        {
            Write-Message "Stopping Monitoring Agent..."
            Stop-MonitoringAgent -MaxWaitSec $StopTaskMaxWaitSec -LogPath $LogDir -SkipIfNotInstalled
        }
        catch
        {
            # Set exit code to indicate Monitoring Agent could not be stopped.
            $ExitCode = 4

            # Throw to terminate script.
            throw
        }
        $ErrorMessages = @()

        # Restarting the EventLog service can prevent some locked file issues.
        Write-Message "Restarting EventLog service..." -Vrb
        try
        {
            Restart-Service -Name "EventLog" -Force
        }
        catch
        {
            Write-Message "Warning: Failed to restart EventLog service: $($_)" -Vrb
        }

        # Get Monitoring manifest path.
        $MonManifestPath = Get-MonitoringManifestPath
        if (!$MonManifestPath)
        {
            $MonManifestPath = Get-MonitoringManifestPathDefault
            Write-Message "Warning: No Monitoring Manifest path found in registry. Using default: $($MonManifestPath)" -Vrb
        }

        Write-Message "Monitoring manifest path: $($MonManifestPath)"

        if (!(Test-Path -Path $MonManifestPath -ErrorAction SilentlyContinue))
        {
            Write-Message "Monitoring manifest path was not found. Creating new directory: $($MonManifestPath)" -Vrb
            New-Item -ItemType Directory -Path $MonManifestPath | Out-Null
        }

        # Set backup path. If the RunbookBackupFolder variable is not set, use ManualETWManifestBackup folder.
        $BackupFolder = $RunbookBackupFolder
        if (!$BackupFolder)
        {
            $BackupFolder = Join-Path -Path $PSScriptRoot -ChildPath "ManualETWManifestBackup"
        }

        if (!(Test-Path -Path $BackupFolder -ErrorAction SilentlyContinue))
        {
            Write-Message "Backup path was not found. Creating new directory: $($BackupFolder)" -Vrb
            New-Item -ItemType Directory -Path $BackupFolder | Out-Null
        }

        # Backup existing files from Monitoring manifest folder prior to update (no subfolders).
        $BackupFiles = @(Get-ChildItem -Path $MonManifestPath -File)
        $BackupErrors = @()
        Write-Message "Backing up $($BackupFiles.Count) files from Monitoring manifest folder..."
        foreach ($File in $BackupFiles)
        {
            $DestinationFilePath = Join-Path -Path $BackupFolder -ChildPath $($File.Name)

            # Only copy file to backup folder if it does not already exist or is zero bytes.
            # In case of resume/retries, the backup may already exist and should not be overwritten.
            if ((Test-Path -Path $DestinationFilePath) -and (Get-Item -Path $DestinationFilePath).Length -gt 0)
            {
                Write-Message "- [Already in backup] $($File.Name)" -Vrb
            }
            else
            {
                Write-Message "- [Adding to backup] $($File.Name)" -Vrb

                [int]$Attempt = 1
                [int]$MaxAttempts = 3
                while ($Attempt -le $MaxAttempts)
                {
                    try
                    {
                        Copy-Item -Path $($File.FullName) -Destination $DestinationFilePath -Force
                    }
                    catch
                    {
                        if ($Attempt -lt $MaxAttempts)
                        {
                            Write-Message "- [Warning - $($Attempt)]: Failed to backup $($File.Name): $($_.Exception.Message)." -Vrb
                            Start-Sleep -Second 1
                        }
                        else
                        {
                            Write-Message "- [Error]: Failed all attempts to backup $($File.FullName): $($_.Exception.Message)." -Vrb
                            $BackupErrors += $File
                        }
                    }

                    $Attempt++
                }
            }
        }
        if ($BackupErrors.Count -gt 0)
        {
            Write-Message "Error: Failed to backup $($BackupErrors.Count) files to backup folder."
            $ErrorMessages += "Failed to backup $($BackupErrors.Count) files to backup folder."
            $ExitCode = $ExitCode -bor 32
        }
        else
        {
            Write-Message "Manifest files were all copied to backup at: $($BackupFolder)"
        }

        # Update Monitoring manifest folder with new files from ETWManifest folder (no subfolders).
        $UpdateFiles = @(Get-ChildItem -Path $ETWManifestPath -File)
        $UpdateErrors = @()
        Write-Message "Deploying $($UpdateFiles.Count) files to Monitoring manifest folder..."
        foreach ($File in $UpdateFiles)
        {
            $DestinationFilePath = Join-Path -Path $MonManifestPath -ChildPath $($File.Name)

            [int]$Attempt = 1
            [int]$MaxAttempts = 3
            while ($Attempt -le $MaxAttempts)
            {
                try
                {
                    if (Test-Path -Path $DestinationFilePath -ErrorAction SilentlyContinue)
                    {
                        Write-Message "- [Updating] $($File.Name)" -Vrb
                        Copy-Item -Path $($File.FullName) -Destination $DestinationFilePath -Force
                    }
                    else
                    {
                        Write-Message "- [Adding] $($File.Name)" -Vrb
                        Copy-Item -Path $($File.FullName) -Destination $DestinationFilePath -Force
                    }
                }
                catch
                {
                    if ($Attempt -lt $MaxAttempts)
                    {
                        Write-Message "- [Warning - $($Attempt)]: Failed to deploy $($File.Name): $($_.Exception.Message)." -Vrb
                        Start-Sleep -Second 1
                    }
                    else
                    {
                        Write-Message "- [Error]: Failed all attempts to deploy $($File.FullName): $($_.Exception.Message)." -Vrb
                        $UpdateErrors += $File
                    }
                }

                $Attempt++
            }
        }

        if ($UpdateErrors.Count -gt 0)
        {
            Write-Message "Error: Failed to deploy $($UpdateErrors.Count) files to Monitoring manifest folder."
            $ErrorMessages += "Failed to deploy $($UpdateErrors.Count) files to Monitoring manifest folder."
            $ExitCode = $ExitCode -bor 32
        }
        else
        {
            Write-Message "Manifest files were all deployed to: $($MonManifestPath)"
        }

        # Start the MonitoringInstall task.
        try
        {
            Write-Message "Starting Monitoring Agent..."
            Start-MonitoringAgent -MaxWaitSec $StartTaskMaxWaitSec -SkipIfNotInstalled
        }
        catch
        {
            Write-Message "Error: Failed to start Monitoring Agent: $($_)"
            $ErrorMessages += "Failed to start Monitoring Agent: $($_)"

            # Set exit code to indicate Monitoring Agent could not be started.
            $ExitCode = $ExitCode -bor 8
        }

        # Throw exception with error messages.
        if ($ErrorMessages.Count -gt 0)
        {
            $ErrorMessage = [string]::Join("; ", $ErrorMessages)
            throw $ErrorMessage
        }
    }
}
catch
{
    # Ensure non-zero exit code if an exception is caught and no exit code set.
    if ($ExitCode -eq 0)
    {
        $ExitCode = 1024
    }

    $ErrorMessage = "Error during ETW update: $($_)"

    Write-Message $ErrorMessage
    Write-Message "$($_.ScriptStackTrace)" -Vrb
    Write-Message "$($_.Exception)" -Vrb
    Write-Message "ETW update script failed with exit code: $($ExitCode)."

    # Use throw to indicate error to AXUpdateInstaller.
    # In case of exceptions, the output is not captured, so only the error message and
    # log file contents will be available for diagnostics.
    throw "$($ErrorMessage) [Log: $($LogPath)]"
}

Write-Message "ETW update script completed with exit code: $($ExitCode)."
exit $ExitCode

# SIG # Begin signature block
# MIIkgAYJKoZIhvcNAQcCoIIkcTCCJG0CAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCCRzUnlJn7hzaFQ
# mMI9ez9Yg6lCbwcyGzu+jgXtR9m0YqCCDYEwggX/MIID56ADAgECAhMzAAABUZ6N
# j0Bxow5BAAAAAAFRMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMTkwNTAyMjEzNzQ2WhcNMjAwNTAyMjEzNzQ2WjB0MQsw
# CQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9u
# ZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMR4wHAYDVQQDExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIB
# AQCVWsaGaUcdNB7xVcNmdfZiVBhYFGcn8KMqxgNIvOZWNH9JYQLuhHhmJ5RWISy1
# oey3zTuxqLbkHAdmbeU8NFMo49Pv71MgIS9IG/EtqwOH7upan+lIq6NOcw5fO6Os
# +12R0Q28MzGn+3y7F2mKDnopVu0sEufy453gxz16M8bAw4+QXuv7+fR9WzRJ2CpU
# 62wQKYiFQMfew6Vh5fuPoXloN3k6+Qlz7zgcT4YRmxzx7jMVpP/uvK6sZcBxQ3Wg
# B/WkyXHgxaY19IAzLq2QiPiX2YryiR5EsYBq35BP7U15DlZtpSs2wIYTkkDBxhPJ
# IDJgowZu5GyhHdqrst3OjkSRAgMBAAGjggF+MIIBejAfBgNVHSUEGDAWBgorBgEE
# AYI3TAgBBggrBgEFBQcDAzAdBgNVHQ4EFgQUV4Iarkq57esagu6FUBb270Zijc8w
# UAYDVR0RBEkwR6RFMEMxKTAnBgNVBAsTIE1pY3Jvc29mdCBPcGVyYXRpb25zIFB1
# ZXJ0byBSaWNvMRYwFAYDVQQFEw0yMzAwMTIrNDU0MTM1MB8GA1UdIwQYMBaAFEhu
# ZOVQBdOCqhc3NyK1bajKdQKVMFQGA1UdHwRNMEswSaBHoEWGQ2h0dHA6Ly93d3cu
# bWljcm9zb2Z0LmNvbS9wa2lvcHMvY3JsL01pY0NvZFNpZ1BDQTIwMTFfMjAxMS0w
# Ny0wOC5jcmwwYQYIKwYBBQUHAQEEVTBTMFEGCCsGAQUFBzAChkVodHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NlcnRzL01pY0NvZFNpZ1BDQTIwMTFfMjAx
# MS0wNy0wOC5jcnQwDAYDVR0TAQH/BAIwADANBgkqhkiG9w0BAQsFAAOCAgEAWg+A
# rS4Anq7KrogslIQnoMHSXUPr/RqOIhJX+32ObuY3MFvdlRElbSsSJxrRy/OCCZdS
# se+f2AqQ+F/2aYwBDmUQbeMB8n0pYLZnOPifqe78RBH2fVZsvXxyfizbHubWWoUf
# NW/FJlZlLXwJmF3BoL8E2p09K3hagwz/otcKtQ1+Q4+DaOYXWleqJrJUsnHs9UiL
# crVF0leL/Q1V5bshob2OTlZq0qzSdrMDLWdhyrUOxnZ+ojZ7UdTY4VnCuogbZ9Zs
# 9syJbg7ZUS9SVgYkowRsWv5jV4lbqTD+tG4FzhOwcRQwdb6A8zp2Nnd+s7VdCuYF
# sGgI41ucD8oxVfcAMjF9YX5N2s4mltkqnUe3/htVrnxKKDAwSYliaux2L7gKw+bD
# 1kEZ/5ozLRnJ3jjDkomTrPctokY/KaZ1qub0NUnmOKH+3xUK/plWJK8BOQYuU7gK
# YH7Yy9WSKNlP7pKj6i417+3Na/frInjnBkKRCJ/eYTvBH+s5guezpfQWtU4bNo/j
# 8Qw2vpTQ9w7flhH78Rmwd319+YTmhv7TcxDbWlyteaj4RK2wk3pY1oSz2JPE5PNu
# Nmd9Gmf6oePZgy7Ii9JLLq8SnULV7b+IP0UXRY9q+GdRjM2AEX6msZvvPCIoG0aY
# HQu9wZsKEK2jqvWi8/xdeeeSI9FN6K1w4oVQM4Mwggd6MIIFYqADAgECAgphDpDS
# AAAAAAADMA0GCSqGSIb3DQEBCwUAMIGIMQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMTIwMAYDVQQDEylNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0
# ZSBBdXRob3JpdHkgMjAxMTAeFw0xMTA3MDgyMDU5MDlaFw0yNjA3MDgyMTA5MDla
# MH4xCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdS
# ZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMT
# H01pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTEwggIiMA0GCSqGSIb3DQEB
# AQUAA4ICDwAwggIKAoICAQCr8PpyEBwurdhuqoIQTTS68rZYIZ9CGypr6VpQqrgG
# OBoESbp/wwwe3TdrxhLYC/A4wpkGsMg51QEUMULTiQ15ZId+lGAkbK+eSZzpaF7S
# 35tTsgosw6/ZqSuuegmv15ZZymAaBelmdugyUiYSL+erCFDPs0S3XdjELgN1q2jz
# y23zOlyhFvRGuuA4ZKxuZDV4pqBjDy3TQJP4494HDdVceaVJKecNvqATd76UPe/7
# 4ytaEB9NViiienLgEjq3SV7Y7e1DkYPZe7J7hhvZPrGMXeiJT4Qa8qEvWeSQOy2u
# M1jFtz7+MtOzAz2xsq+SOH7SnYAs9U5WkSE1JcM5bmR/U7qcD60ZI4TL9LoDho33
# X/DQUr+MlIe8wCF0JV8YKLbMJyg4JZg5SjbPfLGSrhwjp6lm7GEfauEoSZ1fiOIl
# XdMhSz5SxLVXPyQD8NF6Wy/VI+NwXQ9RRnez+ADhvKwCgl/bwBWzvRvUVUvnOaEP
# 6SNJvBi4RHxF5MHDcnrgcuck379GmcXvwhxX24ON7E1JMKerjt/sW5+v/N2wZuLB
# l4F77dbtS+dJKacTKKanfWeA5opieF+yL4TXV5xcv3coKPHtbcMojyyPQDdPweGF
# RInECUzF1KVDL3SV9274eCBYLBNdYJWaPk8zhNqwiBfenk70lrC8RqBsmNLg1oiM
# CwIDAQABo4IB7TCCAekwEAYJKwYBBAGCNxUBBAMCAQAwHQYDVR0OBBYEFEhuZOVQ
# BdOCqhc3NyK1bajKdQKVMBkGCSsGAQQBgjcUAgQMHgoAUwB1AGIAQwBBMAsGA1Ud
# DwQEAwIBhjAPBgNVHRMBAf8EBTADAQH/MB8GA1UdIwQYMBaAFHItOgIxkEO5FAVO
# 4eqnxzHRI4k0MFoGA1UdHwRTMFEwT6BNoEuGSWh0dHA6Ly9jcmwubWljcm9zb2Z0
# LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01pY1Jvb0NlckF1dDIwMTFfMjAxMV8wM18y
# Mi5jcmwwXgYIKwYBBQUHAQEEUjBQME4GCCsGAQUFBzAChkJodHRwOi8vd3d3Lm1p
# Y3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY1Jvb0NlckF1dDIwMTFfMjAxMV8wM18y
# Mi5jcnQwgZ8GA1UdIASBlzCBlDCBkQYJKwYBBAGCNy4DMIGDMD8GCCsGAQUFBwIB
# FjNodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2RvY3MvcHJpbWFyeWNw
# cy5odG0wQAYIKwYBBQUHAgIwNB4yIB0ATABlAGcAYQBsAF8AcABvAGwAaQBjAHkA
# XwBzAHQAYQB0AGUAbQBlAG4AdAAuIB0wDQYJKoZIhvcNAQELBQADggIBAGfyhqWY
# 4FR5Gi7T2HRnIpsLlhHhY5KZQpZ90nkMkMFlXy4sPvjDctFtg/6+P+gKyju/R6mj
# 82nbY78iNaWXXWWEkH2LRlBV2AySfNIaSxzzPEKLUtCw/WvjPgcuKZvmPRul1LUd
# d5Q54ulkyUQ9eHoj8xN9ppB0g430yyYCRirCihC7pKkFDJvtaPpoLpWgKj8qa1hJ
# Yx8JaW5amJbkg/TAj/NGK978O9C9Ne9uJa7lryft0N3zDq+ZKJeYTQ49C/IIidYf
# wzIY4vDFLc5bnrRJOQrGCsLGra7lstnbFYhRRVg4MnEnGn+x9Cf43iw6IGmYslmJ
# aG5vp7d0w0AFBqYBKig+gj8TTWYLwLNN9eGPfxxvFX1Fp3blQCplo8NdUmKGwx1j
# NpeG39rz+PIWoZon4c2ll9DuXWNB41sHnIc+BncG0QaxdR8UvmFhtfDcxhsEvt9B
# xw4o7t5lL+yX9qFcltgA1qFGvVnzl6UJS0gQmYAf0AApxbGbpT9Fdx41xtKiop96
# eiL6SJUfq/tHI4D1nvi/a7dLl+LrdXga7Oo3mXkYS//WsyNodeav+vyL6wuA6mk7
# r/ww7QRMjt/fdW1jkT3RnVZOT7+AVyKheBEyIXrvQQqxP/uozKRdwaGIm1dxVk5I
# RcBCyZt2WwqASGv9eZ/BvW1taslScxMNelDNMYIWVTCCFlECAQEwgZUwfjELMAkG
# A1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQx
# HjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEoMCYGA1UEAxMfTWljcm9z
# b2Z0IENvZGUgU2lnbmluZyBQQ0EgMjAxMQITMwAAAVGejY9AcaMOQQAAAAABUTAN
# BglghkgBZQMEAgEFAKCB1jAZBgkqhkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAcBgor
# BgEEAYI3AgELMQ4wDAYKKwYBBAGCNwIBFTAvBgkqhkiG9w0BCQQxIgQgh93k9YYT
# XlRY71t70XvXD6p8mIbC+STiIUIk5Yec5xUwagYKKwYBBAGCNwIBDDFcMFqgPIA6
# AEEAdQB0AG8ARABlAHAAbABvAHkAUgBlAHAAbwByAHQAQQBuAGQAUwB5AG4AYwBE
# AEIALgBwAHMAMaEagBhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcN
# AQEBBQAEggEAgzkb3Ks1y2alAE7YdI4r/ZrZgGUN7KGWETNGEowX3KuOwJqxqYWT
# jqut+q6YYcQ6iAaaraH8DZHr67xUuqrcZkel523L6rxnnvt+eskomX7dqN2dFVpK
# Kl8DjTVtRRKU3xDcgFkxkeQnSj4i6nWAdapTmP7yr8YOU4YbsfsOke3n1rLgJj0g
# hQcOPURyjGW9PPRt9/sJWpK5QV4ewjZCYZIqE3esWQRNITXLcBy62XLbxgFG0H+n
# vkpKo7BG/rlwg3x4Hk6R2RUJsNs5QKnUta81+yd2/6gf3bfbJ6CkrZ9/Z+mjmYUx
# +CP76+EGU5eYzXMz3c26E7f3lXareBbXgaGCE7cwghOzBgorBgEEAYI3AwMBMYIT
# ozCCE58GCSqGSIb3DQEHAqCCE5AwghOMAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFY
# BgsqhkiG9w0BCRABBKCCAUcEggFDMIIBPwIBAQYKKwYBBAGEWQoDATAxMA0GCWCG
# SAFlAwQCAQUABCDnvxwbX4MorHt/bZveB1i9ZQk8PKSb02M4Bqr6e1lyKwIGXkwK
# ApmaGBMyMDIwMDMyNDA2NTAwOS4zMjRaMAcCAQGAAgH0oIHUpIHRMIHOMQswCQYD
# VQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEe
# MBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSkwJwYDVQQLEyBNaWNyb3Nv
# ZnQgT3BlcmF0aW9ucyBQdWVydG8gUmljbzEmMCQGA1UECxMdVGhhbGVzIFRTUyBF
# U046QzBGNC0zMDg2LURFRjgxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1w
# IFNlcnZpY2Wggg8fMIIE9TCCA92gAwIBAgITMwAAAQGBPiJG/fo44gAAAAABATAN
# BgkqhkiG9w0BAQsFADB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3Rv
# bjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0
# aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMDAeFw0x
# OTA5MDYyMDQxMTVaFw0yMDEyMDQyMDQxMTVaMIHOMQswCQYDVQQGEwJVUzETMBEG
# A1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWlj
# cm9zb2Z0IENvcnBvcmF0aW9uMSkwJwYDVQQLEyBNaWNyb3NvZnQgT3BlcmF0aW9u
# cyBQdWVydG8gUmljbzEmMCQGA1UECxMdVGhhbGVzIFRTUyBFU046QzBGNC0zMDg2
# LURFRjgxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2UwggEi
# MA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCyUPZYyITkNb2qIq1P3Shde+CJ
# lHddvcixvnh4AZKR1/LCJdaOqWyBtseYQ14S6o91+WfXSzmZFy94l67m7IZhULAP
# AL1+1DsPfDa5vPG2nDUKTMaI4nfl/O5hpO1GK0JludOPSh+u2lOvd/uY85FrpiYv
# ghmHfDkt7GUC+xE9jDORTl3c+xQL22uUmu01WP5ueRsGSexagFwJ8kALSIavN8eS
# SpdAj2gXY+L1uh3sdBsJ+pf4rqbfM1VuIZO7PcAt0yQgfv+P6MzAyfb8MIC4EIyH
# /Sntu2JkBswgyTCW+RRoWG3q4zZAges8BU7raHJ1mjvmL1xdB/aLeY6VxK9ZAgMB
# AAGjggEbMIIBFzAdBgNVHQ4EFgQUv76Ymu9N/A6MmGCRLYwVJB25pJEwHwYDVR0j
# BBgwFoAU1WM6XIoxkPNDe3xGG8UzaFqFbVUwVgYDVR0fBE8wTTBLoEmgR4ZFaHR0
# cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMvTWljVGltU3Rh
# UENBXzIwMTAtMDctMDEuY3JsMFoGCCsGAQUFBwEBBE4wTDBKBggrBgEFBQcwAoY+
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNUaW1TdGFQQ0Ff
# MjAxMC0wNy0wMS5jcnQwDAYDVR0TAQH/BAIwADATBgNVHSUEDDAKBggrBgEFBQcD
# CDANBgkqhkiG9w0BAQsFAAOCAQEAHz22Wt6f5Ve2WJBP0Vn1quoEoyCe/4xU4MBK
# QVHn8kxz50BtwRdva6tI6JDqonlKg0YsGF4Olpp5scMn5lbEggEXSV32x1gt8Dic
# NB9Cn1CyJqmdA8QWKA4PE1y1VYDf0IJfIKWM59J/9kDb+5xFVl5aSDbnmdKoRn+x
# obYacuNIffSxQvJpivzvnHig0R2GSGHjTfI2kJJUQS1qkS3dppg0r0cXzeFR3BXv
# 04cfNys71r1I2jpN8xn+dDpkKcro1OBCiDstTypiGyhqyxL2fxL6cS2wYbB0UuLX
# At+z4dt7YeDvsBCv24OeYDR79FPSI8EIy8fM7LuhRVJU32t9YDCCBnEwggRZoAMC
# AQICCmEJgSoAAAAAAAIwDQYJKoZIhvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xMjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENl
# cnRpZmljYXRlIEF1dGhvcml0eSAyMDEwMB4XDTEwMDcwMTIxMzY1NVoXDTI1MDcw
# MTIxNDY1NVowfDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAO
# BgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEm
# MCQGA1UEAxMdTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwggEiMA0GCSqG
# SIb3DQEBAQUAA4IBDwAwggEKAoIBAQCpHQ28dxGKOiDs/BOX9fp/aZRrdFQQ1aUK
# AIKF++18aEssX8XD5WHCdrc+Zitb8BVTJwQxH0EbGpUdzgkTjnxhMFmxMEQP8WCI
# hFRDDNdNuDgIs0Ldk6zWczBXJoKjRQ3Q6vVHgc2/JGAyWGBG8lhHhjKEHnRhZ5Ff
# gVSxz5NMksHEpl3RYRNuKMYa+YaAu99h/EbBJx0kZxJyGiGKr0tkiVBisV39dx89
# 8Fd1rL2KQk1AUdEPnAY+Z3/1ZsADlkR+79BL/W7lmsqxqPJ6Kgox8NpOBpG2iAg1
# 6HgcsOmZzTznL0S6p/TcZL2kAcEgCZN4zfy8wMlEXV4WnAEFTyJNAgMBAAGjggHm
# MIIB4jAQBgkrBgEEAYI3FQEEAwIBADAdBgNVHQ4EFgQU1WM6XIoxkPNDe3xGG8Uz
# aFqFbVUwGQYJKwYBBAGCNxQCBAweCgBTAHUAYgBDAEEwCwYDVR0PBAQDAgGGMA8G
# A1UdEwEB/wQFMAMBAf8wHwYDVR0jBBgwFoAU1fZWy4/oolxiaNE9lJBb186aGMQw
# VgYDVR0fBE8wTTBLoEmgR4ZFaHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9j
# cmwvcHJvZHVjdHMvTWljUm9vQ2VyQXV0XzIwMTAtMDYtMjMuY3JsMFoGCCsGAQUF
# BwEBBE4wTDBKBggrBgEFBQcwAoY+aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3Br
# aS9jZXJ0cy9NaWNSb29DZXJBdXRfMjAxMC0wNi0yMy5jcnQwgaAGA1UdIAEB/wSB
# lTCBkjCBjwYJKwYBBAGCNy4DMIGBMD0GCCsGAQUFBwIBFjFodHRwOi8vd3d3Lm1p
# Y3Jvc29mdC5jb20vUEtJL2RvY3MvQ1BTL2RlZmF1bHQuaHRtMEAGCCsGAQUFBwIC
# MDQeMiAdAEwAZQBnAGEAbABfAFAAbwBsAGkAYwB5AF8AUwB0AGEAdABlAG0AZQBu
# AHQALiAdMA0GCSqGSIb3DQEBCwUAA4ICAQAH5ohRDeLG4Jg/gXEDPZ2joSFvs+um
# zPUxvs8F4qn++ldtGTCzwsVmyWrf9efweL3HqJ4l4/m87WtUVwgrUYJEEvu5U4zM
# 9GASinbMQEBBm9xcF/9c+V4XNZgkVkt070IQyK+/f8Z/8jd9Wj8c8pl5SpFSAK84
# Dxf1L3mBZdmptWvkx872ynoAb0swRCQiPM/tA6WWj1kpvLb9BOFwnzJKJ/1Vry/+
# tuWOM7tiX5rbV0Dp8c6ZZpCM/2pif93FSguRJuI57BlKcWOdeyFtw5yjojz6f32W
# apB4pm3S4Zz5Hfw42JT0xqUKloakvZ4argRCg7i1gJsiOCC1JeVk7Pf0v35jWSUP
# ei45V3aicaoGig+JFrphpxHLmtgOR5qAxdDNp9DvfYPw4TtxCd9ddJgiCGHasFAe
# b73x4QDf5zEHpJM692VHeOj4qEir995yfmFrb3epgcunCaw5u+zGy9iCtHLNHfS4
# hQEegPsbiSpUObJb2sgNVZl6h3M7COaYLeqN4DMuEin1wC9UJyH3yKxO2ii4sanb
# lrKnQqLJzxlBTeCG+SqaoxFmMNO7dDJL32N79ZmKLxvHIa9Zta7cRDyXUHHXodLF
# VeNp3lfB0d4wwP3M5k37Db9dT+mdHhk4L7zPWAUu7w2gUDXa7wknHNWzfjUeCLra
# NtvTX4/edIhJEqGCA60wggKVAgEBMIH+oYHUpIHRMIHOMQswCQYDVQQGEwJVUzET
# MBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMV
# TWljcm9zb2Z0IENvcnBvcmF0aW9uMSkwJwYDVQQLEyBNaWNyb3NvZnQgT3BlcmF0
# aW9ucyBQdWVydG8gUmljbzEmMCQGA1UECxMdVGhhbGVzIFRTUyBFU046QzBGNC0z
# MDg2LURFRjgxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2Wi
# JQoBATAJBgUrDgMCGgUAAxUACCVTQOIggfoN27b+VXzd4SffBfGggd4wgdukgdgw
# gdUxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdS
# ZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xKTAnBgNVBAsT
# IE1pY3Jvc29mdCBPcGVyYXRpb25zIFB1ZXJ0byBSaWNvMScwJQYDVQQLEx5uQ2lw
# aGVyIE5UUyBFU046NERFOS0wQzVFLTNFMDkxKzApBgNVBAMTIk1pY3Jvc29mdCBU
# aW1lIFNvdXJjZSBNYXN0ZXIgQ2xvY2swDQYJKoZIhvcNAQEFBQACBQDiI5+DMCIY
# DzIwMjAwMzI0MDQ1MDExWhgPMjAyMDAzMjUwNDUwMTFaMHQwOgYKKwYBBAGEWQoE
# ATEsMCowCgIFAOIjn4MCAQAwBwIBAAICBXQwBwIBAAICGv8wCgIFAOIk8QMCAQAw
# NgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGEWQoDAaAKMAgCAQACAxbjYKEKMAgC
# AQACAwehIDANBgkqhkiG9w0BAQUFAAOCAQEAVCiukCUy4YF1vnMnk4rxsOzfADZ2
# YXDETuia5qoYmaI7zj2L+lKl2tTRs8PA7el3EkQ+Pavg+q6TO7YIFwzUsnQdB6kA
# XC0nMDyTwohx5MYj/ZnosX8ia9lyBKtbzgTfObGhHGgbkO5eutGM1pxawQy7mwDW
# VgXrkkxmqCZ1c82yEbN652JaZkwdO7eLYjiKheTsYuFcsrXGdydSVa6wBtXKCQ+s
# 09IYqTakfFCYnudCxamLH3yLCNm4xcIEub7AG2ZyV1jUjPpk40dIwOI8L2FKa6o7
# lZOCmMuzA2u1YmopuxDo53QLeTg+kSargaeEQc9nhOiOd0iCCiojyDalGjGCAvUw
# ggLxAgEBMIGTMHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAw
# DgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24x
# JjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAyMDEwAhMzAAABAYE+
# Ikb9+jjiAAAAAAEBMA0GCWCGSAFlAwQCAQUAoIIBMjAaBgkqhkiG9w0BCQMxDQYL
# KoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkEMSIEIExl2N1WBwfQRU9PS4eZx5pTUW7G
# fTMZNZXnSCRkpMhFMIHiBgsqhkiG9w0BCRACDDGB0jCBzzCBzDCBsQQUCCVTQOIg
# gfoN27b+VXzd4SffBfEwgZgwgYCkfjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0Eg
# MjAxMAITMwAAAQGBPiJG/fo44gAAAAABATAWBBR2u5sQaRgoeD0UBTWCpuLNPYlT
# tDANBgkqhkiG9w0BAQsFAASCAQCr+m01r59C+5wQPB/7QvcRM2gW7iZAQoXSE19v
# 6PEAUQ9UkIjWNl4em2hj3Qb5fj5OBz8mQnaLYbd3KevnZ3ZoqzTg9qeeA0Jlbik+
# 5LFkdZCi/gIHytlg30hMR0fG5v4ou7eNfuHMJSaCB1skGa3DqdeFdJL59XK+AzpZ
# +PZYRilD9MQeJXlbjkGKAbPstXrZshQl4KDmd0THTyFNIaPhWgZkHGAUSfobZIOL
# SnoEBVu1HjJopeBP0tco27d0zCK6wmlinqo0KDBmzqvB6GIyAn0xXVjTVTTg14ko
# PNzsUcs707QxdbB/mBeBno8G5nGtKVsUea75orjtdvVQFhtd
# SIG # End signature block
