<#
.SYNOPSIS
    Functions to manage the Monitoring Agent during servicing.

.DESCRIPTION
    Contains a set of functions used by the Dynamics 365 for Finance and Operations
    Enterprise Edition servicing process to interact with the Geneva Monitoring Agent.

.NOTES
    Requires PowerShell 3.0 or later.

    Configure Write-Message by calling Set-WriteMessageLogFilePath to set a
    file to send output to.

    Copyright © 2019 Microsoft. All rights reserved.
#>

# Module level variables to control default behavior of Write-Message.
[string]$Script:WriteMessageLogFilePath = $null

<#
.SYNOPSIS
    Set module level variable to control default behavior of Write-Message.
    Set to the full path of the file to append output to (default is null).
#>
function Set-WriteMessageLogFilePath
{
    [CmdletBinding()]
    Param([string]$Path)

    $Script:WriteMessageLogFilePath = $Path
}

<#
.SYNOPSIS
    Write messages to log file and/or output.
#>
function Write-Message
{
    [CmdletBinding()]
    Param(
        [parameter(Mandatory = $false, ValueFromPipeline = $true)][string[]]$Message,
        [switch]$Vrb,
        [string]$LogFilePath = $Script:WriteMessageLogFilePath)

    Process
    {
        if ($LogFilePath)
        {
            $Message | ForEach-Object { "$([DateTime]::UtcNow.ToString("yyyy-MM-dd HH:mm:ss")): $($_)" | Out-File -FilePath $LogFilePath -Append }
        }

        # Verbose messages only goes to log file.
        if (!$Vrb)
        {
            $Message | Write-Output
        }
    }
}

<#
.SYNOPSIS
    Returns the registry path for the MonitoringInstall settings.
#>
function Get-MonitoringInstallRegistryPath
{
    return "HKLM:\SOFTWARE\Microsoft\Dynamics\AX\Diagnostics\MonitoringInstall"
}

<#
.SYNOPSIS
    Returns the scheduled task path for the MonitoringInstall tasks.
#>
function Get-MonitoringInstallTaskPath
{
    # Task path must begin and end with "\" for Get-ScheduledTask to work.
    return "\Microsoft\Dynamics\AX\Diagnostics\"
}

<#
.SYNOPSIS
    Read the Monitoring installation path from registry.
#>
function Get-MonitoringInstallPath()
{
    $RegistryPath = Get-MonitoringInstallRegistryPath
    $InstallPathKey = "InstallPath"
    $InstallPath = $null

    if (Test-Path -LiteralPath $RegistryPath -ErrorAction SilentlyContinue)
    {
        $RegistryKey = Get-Item -LiteralPath $RegistryPath -ErrorAction SilentlyContinue
        if ($RegistryKey)
        {
            $InstallPath = $RegistryKey.GetValue($InstallPathKey)
            if (!$InstallPath)
            {
                Write-Message "Warning: Could not read MonitoringInstall registry key $($InstallPathKey) from: $($RegistryPath)" -Vrb
            }
        }
        else
        {
            Write-Message "Warning: Could not read MonitoringInstall registry keys from: $($RegistryPath)" -Vrb
        }
    }
    else
    {
        Write-Message "Warning: Could not find MonitoringInstall registry path: $($RegistryPath)" -Vrb
    }

    return $InstallPath
}

<#
.SYNOPSIS
    Get the default Monitoring installation path.
#>
function Get-MonitoringInstallPathDefault()
{
    if ($env:SERVICEDRIVE)
    {
        $InstallPath = Join-Path -Path "$($env:SERVICEDRIVE)" -ChildPath "Monitoring"
    }
    else
    {
        $InstallPath = Join-Path -Path "$($env:SystemDrive)" -ChildPath "Monitoring"
    }

    return $InstallPath
}

<#
.SYNOPSIS
    Read the Monitoring manifest path from registry.
#>
function Get-MonitoringManifestPath()
{
    $RegistryPath = Get-MonitoringInstallRegistryPath
    $ManifestPathKey = "ManifestPath"
    $ManifestPath = $null

    if (Test-Path -LiteralPath $RegistryPath -ErrorAction SilentlyContinue)
    {
        $RegistryKey = Get-Item -LiteralPath $RegistryPath -ErrorAction SilentlyContinue
        if ($RegistryKey)
        {
            $ManifestPath = $RegistryKey.GetValue($ManifestPathKey)
            if (!$ManifestPath)
            {
                Write-Message "Warning: Could not read MonitoringInstall registry key $($ManifestPathKey) from: $($RegistryPath)" -Vrb
            }
        }
        else
        {
            Write-Message "Warning: Could not read MonitoringInstall registry key: $($RegistryPath)" -Vrb
        }
    }
    else
    {
        Write-Message "Warning: Could not find MonitoringInstall registry path: $($RegistryPath)" -Vrb
    }

    return $ManifestPath
}

<#
.SYNOPSIS
    Get the default Monitoring manifest path.
#>
function Get-MonitoringManifestPathDefault()
{
    $InstallPath = Get-MonitoringInstallPath
    if (!$InstallPath)
    {
        $InstallPath = Get-MonitoringInstallPathDefault
    }

    $ManifestPath = Join-Path -Path $InstallPath -ChildPath "Instrumentation"

    return $ManifestPath
}

<#
.SYNOPSIS
    Create task for MonitoringInstall.exe action and run as SYSTEM.
#>
function Invoke-MonitoringInstallAsSystem([string]$MonitoringInstallFilePath, [string]$LogPath, [string]$Action, [int]$MaxWaitSec = 300, [int]$MaxLogLines = 100, [switch]$UseExistingTask)
{
    try
    {
        $TaskPath = Get-MonitoringInstallTaskPath
        $TaskName = "MonitoringInstall-$($Action)"
        $MonitoringInstallPath = Split-Path -Path $MonitoringInstallFilePath -Parent

        if ($LogPath)
        {
            if (!(Test-Path -Path $LogPath))
            {
                New-Item -Path $LogPath -ItemType Directory | Out-Null
            }

            $LogFilePath = Join-Path -Path $LogPath -ChildPath "$($TaskName)_$([DateTime]::UtcNow.ToString("yyyyMMddHHmmss")).log"
        }
        else
        {
            # Default to using the log file path in the MonitoringInstall folder.
            $LogFilePath = Join-Path -Path $MonitoringInstallPath -ChildPath "$($TaskName).log"
        }

        Write-Message "Getting the $($TaskName) task..." -Vrb
        $ExistingTask = Get-ScheduledTask -TaskPath $TaskPath -TaskName $TaskName -ErrorAction SilentlyContinue
        if ($ExistingTask)
        {
            if ($UseExistingTask)
            {
                Write-Message "Using existing $($TaskName) task found (State: $($ExistingTask.State))." -Vrb
                $Task = $ExistingTask
            }
            else
            {
                Write-Message "Existing $($TaskName) task found, but it will be replaced (State: $($ExistingTask.State))." -Vrb
            }
        }

        if (!$Task)
        {
            Write-Message "Creating $($TaskName) task to run as SYSTEM." -Vrb
            $TaskAction = New-ScheduledTaskAction -Execute $MonitoringInstallFilePath -Argument "/$($Action) /id:SingleAgent /log:$LogFilePath" -WorkingDirectory $MonitoringInstallPath
            $TaskSettings = New-ScheduledTaskSettingsSet -ExecutionTimeLimit (New-TimeSpan -Seconds $MaxWaitSec)
            $Task = Register-ScheduledTask -TaskPath $TaskPath -TaskName $TaskName -User "SYSTEM" -Action $TaskAction -Settings $TaskSettings -Force
        }

        Write-Message "Starting the $($TaskName) task..." -Vrb
        $Task | Start-ScheduledTask
        $ScheduledTask = $Task | Get-ScheduledTask
        Write-Message "Scheduled task started (State: $($ScheduledTask.State))." -Vrb

        Write-Message "Waiting up to $($MaxWaitSec) seconds for the $($TaskName) task to complete..." -Vrb
        [int]$WaitedSec = 0
        while ($ScheduledTask.State -ine "Ready" -and $WaitedSec -lt $MaxWaitSec)
        {
            Start-Sleep -Seconds 1
            $WaitedSec += 1
            $ScheduledTask = $Task | Get-ScheduledTask
        }

        # Check if task completed.
        if ($ScheduledTask.State -ine "Ready")
        {
            $ErrorMessage = "Exceeded timeout after waiting $($WaitedSec) seconds for the $($TaskName) task to complete (State: $($ScheduledTask.State))."
            Write-Message "Error: $($ErrorMessage)" -Vrb
            throw $ErrorMessage
        }
        else
        {
            Write-Message "Completed the $($TaskName) task after waiting $($WaitedSec) seconds (State: $($ScheduledTask.State))." -Vrb
            $TaskInfo = $ScheduledTask | Get-ScheduledTaskInfo
            if ($TaskInfo)
            {
                # Note: Currently the MonitoringInstall exit code cannot be used to determine success or failure.
                Write-Message "The exit code of the $($TaskName) task run at $($TaskInfo.LastRunTime.ToString("u")) is: $($TaskInfo.LastTaskResult)." -Vrb
            }
            else
            {
                Write-Message "Warning: No task run information could be found from the $($TaskName) task." -Vrb
            }
        }

        # Check log content.
        if (Test-Path -Path $LogFilePath -ErrorAction SilentlyContinue)
        {
            $LogContent = Get-Content -Path $LogFilePath -Tail $MaxLogLines
            if ($LogContent)
            {
                Write-Message "The MonitoringInstall $($Action) action log file exists at: $($LogFilePath)" -Vrb
                Write-Message "Contents of log (Last $($MaxLogLines)):" -Vrb
                foreach ($Line in $LogContent)
                {
                    Write-Message $Line -Vrb
                }

                Write-Message "*** End of Log ***" -Vrb
            }
            else
            {
                Write-Message "Warning: The MonitoringInstall $($Action) action did not write any contents to log file at: $($LogFilePath)" -Vrb
            }
        }
        else
        {
            Write-Message "Warning: The MonitoringInstall $($Action) action did not produce any log file at: $($LogFilePath)" -Vrb
        }
    }
    catch
    {
        Write-Message "Error: Failed to run MonitoringInstall $($Action) action: $($_.Message)." -Vrb
        throw
    }
    finally
    {
        # If task was triggered, make sure it is stopped.
        if ($ScheduledTask)
        {
            $ScheduledTask = $ScheduledTask | Get-ScheduledTask

            if ($ScheduledTask -and $ScheduledTask.State -ieq "Running")
            {
                Write-Message "Stopping the $($ScheduledTask.TaskName) task as it is still running." -Vrb
                $ScheduledTask | Stop-ScheduledTask
            }
        }
    }
}

<#
.SYNOPSIS
    Stop Monitoring Agent ETW sessions and processes.
#>
function Stop-MonitoringAgent([int]$MaxWaitSec = 300, [string]$LogPath, [switch]$SkipIfNotInstalled)
{
    $InstallPath = Get-MonitoringInstallPath
    if (!$InstallPath)
    {
        # Return without action if Monitoring Install path was not found in registry and skip switch specified.
        if ($SkipIfNotInstalled)
        {
            Write-Message "No Monitoring Install path found in registry. No attempt to stop the Monitoring Agent will be made." -Vrb
            return
        }
        else
        {
            $InstallPath = Get-MonitoringInstallPathDefault
            Write-Message "Warning: No Monitoring Install path found in registry. Using default: $($InstallPath)." -Vrb
        }
    }

    $MonitoringInstallPath = Join-Path -Path $InstallPath -ChildPath "MonitoringInstall"
    if (!(Test-Path -Path $MonitoringInstallPath -ErrorAction SilentlyContinue))
    {
        throw "No MonitoringInstall folder found at: $($MonitoringInstallPath)"
    }

    $MonitoringInstallFilePath = Join-Path -Path $MonitoringInstallPath -ChildPath "MonitoringInstall.exe"
    if (!(Test-Path -Path $MonitoringInstallFilePath -ErrorAction SilentlyContinue))
    {
        throw "No MonitoringInstall.exe file found at: $($MonitoringInstallFilePath)"
    }

    # Stop the Monitoring Install task if it is running.
    # This must be done before attempting to stop Monitoring Agent as it could otherwise
    # be triggering a start of Monitoring Agent after attempting to stop it.
    try
    {
        Write-Message "Stopping Monitoring Install task..."
        Stop-MonitoringInstallTask -MaxWaitSec $MaxWaitSec
    }
    catch
    {
        # This may not be a problem that requires the update process to stop.
        Write-Message "Warning: Failed to stop the MonitoringInstall task: $($_.Message)."
    }

    # Stop Monitoring ETW Sessions as SYSTEM.
    try
    {
        Write-Message "Stopping Monitoring ETW Sessions..."
        Invoke-MonitoringInstallAsSystem -MonitoringInstallFilePath $MonitoringInstallFilePath -Action "StopSessions" -LogPath $LogPath
    }
    catch
    {
        # This may not be a problem that requires the update process to stop.
        Write-Message "Warning: MonitoringInstall failed to stop Monitoring ETW Sessions: $($_.Message)."
    }

    # Stop Monitoring Agent processes as SYSTEM.
    try
    {
        Write-Message "Stopping Monitoring Agent processes..."

        $StopAgentAction = "StopAgents"

        # Determine if MonitoringInstall.exe has the new and improved StopAgentLauncher action.
        # First released with MonitoringInstall.exe 9.3.1789.0 in January 2019.
        # Note: The old StopAgent doesn't actually stop the MonAgentLauncher process, so the other
        # Monitoring Agent processes will be restarted if it does not forcefully terminate it in time.
        $MonitoringInstallFile = Get-Item -Path $MonitoringInstallFilePath
        if ($MonitoringInstallFile.VersionInfo.FileVersion)
        {
            if ($MonitoringInstallFile.VersionInfo.FileMajorPart -ge 9 -and $MonitoringInstallFile.VersionInfo.FileMinorPart -ge 3 -and $MonitoringInstallFile.VersionInfo.FileBuildPart -ge 1789)
            {
                Write-Message "Using MonitoringInstall.exe $($MonitoringInstallFile.VersionInfo.FileVersion) with StopAgentLauncher action." -Vrb
                $StopAgentAction = "StopAgentLauncher"
            }
            else
            {
                Write-Message "Using MonitoringInstall.exe $($MonitoringInstallFile.VersionInfo.FileVersion) with StopAgents action." -Vrb
            }
        }
        else
        {
            Write-Message "Using MonitoringInstall.exe which has no file version information with StopAgents action." -Vrb
        }

        Invoke-MonitoringInstallAsSystem -MonitoringInstallFilePath $MonitoringInstallFilePath -Action $StopAgentAction -LogPath $LogPath
    }
    catch
    {
        # Not being able to stop the Monitoring Agent processes is a fatal problem.
        Write-Message "Error: MonitoringInstall failed to stop Monitoring Agent processes: $($_.Message)." -Vrb
        throw
    }
    finally
    {
        # Log any running Monitoring Agent processes after stop.
        try
        {
            $Processes = Get-MonitoringAgentProcesses
            if ($Processes.Count -gt 0)
            {
                Write-Message "Found $($Processes.Count) Monitoring Agent processes running after stop:" -Vrb
                foreach ($Process in $Processes)
                {
                    Write-Message "- $($Process.Id): $($Process.Name) started at $($Process.StartTime.Tostring("u"))." -Vrb
                }
            }
            else
            {
                Write-Message "Found no Monitoring Agent processes running after stop." -Vrb
            }
        }
        catch
        {
            Write-Message "Warning: Unable to get Monitoring Agent processes running: ($_)" -Vrb
        }
    }
}

<#
.SYNOPSIS
    Start the task to register ETW events, build configurations, and start the Monitoring Agent.
#>
function Start-MonitoringAgent([int]$MaxWaitSec = 300, [switch]$SkipIfNotInstalled)
{
    $TaskPath = Get-MonitoringInstallTaskPath
    $TaskName = "MonitoringInstall"

    Write-Message "Getting the $($TaskName) task..." -Vrb
    $Task = Get-ScheduledTask -TaskPath $TaskPath -TaskName $TaskName -ErrorAction SilentlyContinue
    if (!$Task)
    {
        # Very old installations created the task in the root path.
        Write-Message "Warning: No '$($TaskName)' task found in path '$($TaskPath)'. Looking for task in all paths..." -Vrb
        $Task = Get-ScheduledTask | Where-Object -FilterScript { $_.TaskName -ieq $TaskName } | Select-Object -First 1
    }

    If ($Task)
    {
        Write-Message "The $($TaskName) task was found (State: $($Task.State))." -Vrb
        Write-Message "Starting the $($TaskName) task..." -Vrb
        $Task | Start-ScheduledTask
        $ScheduledTask = $Task | Get-ScheduledTask
        Write-Message "Scheduled task started (State: $($ScheduledTask.State))." -Vrb

        if ($MaxWaitSec -gt 0)
        {
            Write-Message "Waiting up to $($MaxWaitSec) seconds for the $($TaskName) task to complete..." -Vrb

            [int]$WaitedSec = 0
            while ($ScheduledTask.State -ine "Ready" -and $WaitedSec -lt $MaxWaitSec)
            {
                Start-Sleep -Seconds 1
                $WaitedSec += 1
                $ScheduledTask = $Task | Get-ScheduledTask
            }

            # Log any running Monitoring Agent processes after start.
            try
            {
                $Processes = Get-MonitoringAgentProcesses
                if ($Processes.Count -gt 0)
                {
                    Write-Message "Found $($Processes.Count) Monitoring Agent processes running after start:" -Vrb
                    foreach ($Process in $Processes)
                    {
                        Write-Message "- $($Process.Id): $($Process.Name) started at $($Process.StartTime.Tostring("u"))." -Vrb
                    }
                }
                else
                {
                    Write-Message "Found no Monitoring Agent processes running after start." -Vrb
                }
            }
            catch
            {
                Write-Message "Warning: Unable to get Monitoring Agent processes running: ($_)" -Vrb
            }

            # Check if task completed.
            if ($ScheduledTask.State -ine "Ready")
            {
                $ErrorMessage = "Exceeded timeout after waiting $($WaitedSec) seconds for the $($TaskName) task to complete (State: $($ScheduledTask.State))."
                Write-Message "Error: $($ErrorMessage)" -Vrb
                throw $ErrorMessage
            }
            else
            {
                Write-Message "Completed the $($TaskName) task after waiting $($WaitedSec) seconds (State: $($ScheduledTask.State))." -Vrb
                $TaskInfo = $ScheduledTask | Get-ScheduledTaskInfo
                if ($TaskInfo)
                {
                    # Note: Currently the MonitoringInstall exit code cannot be used to determine success or failure.
                    Write-Message "The exit code of the $($TaskName) task run at $($TaskInfo.LastRunTime.ToString("u")) is: $($TaskInfo.LastTaskResult)." -Vrb
                }
                else
                {
                    Write-Message "Warning: No task run information could be found from the $($TaskName) task." -Vrb
                }
            }
        }
        else
        {
            Write-Message "Not waiting for the $($TaskName) task to complete." -Vrb
        }
    }
    else
    {
        # Only throw if Monitoring Install task was not found and no skip switch was specified.
        if ($SkipIfNotInstalled)
        {
            Write-Message "No Monitoring Install task found. No attempt to start the Monitoring Agent will be made." -Vrb
        }
        else
        {
            # Throw if the scheduled task is not found.
            throw "No scheduled task with the name '$($TaskName)' was found. Unable to start Monitoring Agent."
        }
    }
}

<#
.SYNOPSIS
    Stop the Monitoring Install task if it is currently running.
#>
function Stop-MonitoringInstallTask([int]$MaxWaitSec = 300)
{
    $TaskPath = Get-MonitoringInstallTaskPath
    $TaskName = "MonitoringInstall"

    Write-Message "Getting the $($TaskName) task..." -Vrb
    $Task = Get-ScheduledTask -TaskPath $TaskPath -TaskName $TaskName -ErrorAction SilentlyContinue
    if (!$Task)
    {
        # Very old installations created the task in the root path.
        Write-Message "Warning: No '$($TaskName)' task found in path '$($TaskPath)'. Looking for task in all paths..." -Vrb
        $Task = Get-ScheduledTask | Where-Object -FilterScript { $_.TaskName -ieq $TaskName } | Select-Object -First 1
    }

    If ($Task)
    {
        Write-Message "The $($TaskName) task was found (State: $($Task.State))." -Vrb
        if ($Task.State -ieq "Running")
        {
            Write-Message "Stopping the running $($TaskName) task..." -Vrb
            $Task | Stop-ScheduledTask
            $ScheduledTask = $Task | Get-ScheduledTask
            Write-Message "Scheduled task stopping (State: $($ScheduledTask.State))." -Vrb

            if ($MaxWaitSec -gt 0)
            {
                Write-Message "Waiting up to $($MaxWaitSec) seconds for the $($TaskName) task to stop..." -Vrb

                [int]$WaitedSec = 0
                while ($ScheduledTask.State -ine "Ready" -and $WaitedSec -lt $MaxWaitSec)
                {
                    Start-Sleep -Seconds 1
                    $WaitedSec += 1
                    $ScheduledTask = $Task | Get-ScheduledTask
                }

                # Check if task completed.
                if ($ScheduledTask.State -ine "Ready")
                {
                    $ErrorMessage = "Exceeded timeout after waiting $($WaitedSec) seconds for the $($TaskName) task to stop (State: $($ScheduledTask.State))."
                    Write-Message "Error: $($ErrorMessage)" -Vrb
                    throw $ErrorMessage
                }
                else
                {
                    Write-Message "Stopped the $($TaskName) task after waiting $($WaitedSec) seconds (State: $($ScheduledTask.State))." -Vrb

                    # Log task info only as additional information. The objective of this function is to stop
                    # the task if it is running, so the result of the task is not important.
                    $TaskInfo = $ScheduledTask | Get-ScheduledTaskInfo -ErrorAction SilentlyContinue
                    if ($TaskInfo)
                    {
                        Write-Message "The exit code of the $($TaskName) task run at $($TaskInfo.LastRunTime.ToString("u")) is: $($TaskInfo.LastTaskResult)." -Vrb
                    }
                    else
                    {
                        Write-Message "Warning: No task run information could be found from the $($TaskName) task." -Vrb
                    }
                }
            }
            else
            {
                Write-Message "Not waiting for the $($TaskName) task to stop." -Vrb
            }
        }
        else
        {
            Write-Message "Not stopping the $($TaskName) task as it is not running." -Vrb
        }
    }
    else
    {
        # If there is no task found, it is not running, and there is no action required to stop it.
        Write-Message "No Monitoring Install task was found, so there is no task to stop." -Vrb
    }
}

<#
.SYNOPSIS
    Returns the names of the Monitoring Agent processes.
#>
function Get-MonitoringAgentProcessNames
{
    return @("MonAgentLauncher", "MonAgentHost", "MonAgentManager", "MonAgentCore");
}

<#
.SYNOPSIS
    Get all Monitoring Agent processes currently running.
#>
function Get-MonitoringAgentProcesses()
{
    $AgentProcesses = @()

    $AgentProcessNames = Get-MonitoringAgentProcessNames
    foreach ($AgentProcessName in $AgentProcessNames)
    {
        $AgentProcess = Get-Process -Name $AgentProcessName -ErrorAction SilentlyContinue
        if ($AgentProcess)
        {
            if (!$AgentProcess.HasExited)
            {
                $AgentProcesses += $AgentProcess
            }
        }
    }

    return $AgentProcesses
}
# SIG # Begin signature block
# MIIkgwYJKoZIhvcNAQcCoIIkdDCCJHACAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCAMYUm4B3753drr
# QK+APSrQIgRuDo+vilvZIQcra7KgX6CCDYEwggX/MIID56ADAgECAhMzAAABUZ6N
# j0Bxow5BAAAAAAFRMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMTkwNTAyMjEzNzQ2WhcNMjAwNTAyMjEzNzQ2WjB0MQsw
# CQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9u
# ZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMR4wHAYDVQQDExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIB
# AQCVWsaGaUcdNB7xVcNmdfZiVBhYFGcn8KMqxgNIvOZWNH9JYQLuhHhmJ5RWISy1
# oey3zTuxqLbkHAdmbeU8NFMo49Pv71MgIS9IG/EtqwOH7upan+lIq6NOcw5fO6Os
# +12R0Q28MzGn+3y7F2mKDnopVu0sEufy453gxz16M8bAw4+QXuv7+fR9WzRJ2CpU
# 62wQKYiFQMfew6Vh5fuPoXloN3k6+Qlz7zgcT4YRmxzx7jMVpP/uvK6sZcBxQ3Wg
# B/WkyXHgxaY19IAzLq2QiPiX2YryiR5EsYBq35BP7U15DlZtpSs2wIYTkkDBxhPJ
# IDJgowZu5GyhHdqrst3OjkSRAgMBAAGjggF+MIIBejAfBgNVHSUEGDAWBgorBgEE
# AYI3TAgBBggrBgEFBQcDAzAdBgNVHQ4EFgQUV4Iarkq57esagu6FUBb270Zijc8w
# UAYDVR0RBEkwR6RFMEMxKTAnBgNVBAsTIE1pY3Jvc29mdCBPcGVyYXRpb25zIFB1
# ZXJ0byBSaWNvMRYwFAYDVQQFEw0yMzAwMTIrNDU0MTM1MB8GA1UdIwQYMBaAFEhu
# ZOVQBdOCqhc3NyK1bajKdQKVMFQGA1UdHwRNMEswSaBHoEWGQ2h0dHA6Ly93d3cu
# bWljcm9zb2Z0LmNvbS9wa2lvcHMvY3JsL01pY0NvZFNpZ1BDQTIwMTFfMjAxMS0w
# Ny0wOC5jcmwwYQYIKwYBBQUHAQEEVTBTMFEGCCsGAQUFBzAChkVodHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NlcnRzL01pY0NvZFNpZ1BDQTIwMTFfMjAx
# MS0wNy0wOC5jcnQwDAYDVR0TAQH/BAIwADANBgkqhkiG9w0BAQsFAAOCAgEAWg+A
# rS4Anq7KrogslIQnoMHSXUPr/RqOIhJX+32ObuY3MFvdlRElbSsSJxrRy/OCCZdS
# se+f2AqQ+F/2aYwBDmUQbeMB8n0pYLZnOPifqe78RBH2fVZsvXxyfizbHubWWoUf
# NW/FJlZlLXwJmF3BoL8E2p09K3hagwz/otcKtQ1+Q4+DaOYXWleqJrJUsnHs9UiL
# crVF0leL/Q1V5bshob2OTlZq0qzSdrMDLWdhyrUOxnZ+ojZ7UdTY4VnCuogbZ9Zs
# 9syJbg7ZUS9SVgYkowRsWv5jV4lbqTD+tG4FzhOwcRQwdb6A8zp2Nnd+s7VdCuYF
# sGgI41ucD8oxVfcAMjF9YX5N2s4mltkqnUe3/htVrnxKKDAwSYliaux2L7gKw+bD
# 1kEZ/5ozLRnJ3jjDkomTrPctokY/KaZ1qub0NUnmOKH+3xUK/plWJK8BOQYuU7gK
# YH7Yy9WSKNlP7pKj6i417+3Na/frInjnBkKRCJ/eYTvBH+s5guezpfQWtU4bNo/j
# 8Qw2vpTQ9w7flhH78Rmwd319+YTmhv7TcxDbWlyteaj4RK2wk3pY1oSz2JPE5PNu
# Nmd9Gmf6oePZgy7Ii9JLLq8SnULV7b+IP0UXRY9q+GdRjM2AEX6msZvvPCIoG0aY
# HQu9wZsKEK2jqvWi8/xdeeeSI9FN6K1w4oVQM4Mwggd6MIIFYqADAgECAgphDpDS
# AAAAAAADMA0GCSqGSIb3DQEBCwUAMIGIMQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMTIwMAYDVQQDEylNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0
# ZSBBdXRob3JpdHkgMjAxMTAeFw0xMTA3MDgyMDU5MDlaFw0yNjA3MDgyMTA5MDla
# MH4xCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdS
# ZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMT
# H01pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTEwggIiMA0GCSqGSIb3DQEB
# AQUAA4ICDwAwggIKAoICAQCr8PpyEBwurdhuqoIQTTS68rZYIZ9CGypr6VpQqrgG
# OBoESbp/wwwe3TdrxhLYC/A4wpkGsMg51QEUMULTiQ15ZId+lGAkbK+eSZzpaF7S
# 35tTsgosw6/ZqSuuegmv15ZZymAaBelmdugyUiYSL+erCFDPs0S3XdjELgN1q2jz
# y23zOlyhFvRGuuA4ZKxuZDV4pqBjDy3TQJP4494HDdVceaVJKecNvqATd76UPe/7
# 4ytaEB9NViiienLgEjq3SV7Y7e1DkYPZe7J7hhvZPrGMXeiJT4Qa8qEvWeSQOy2u
# M1jFtz7+MtOzAz2xsq+SOH7SnYAs9U5WkSE1JcM5bmR/U7qcD60ZI4TL9LoDho33
# X/DQUr+MlIe8wCF0JV8YKLbMJyg4JZg5SjbPfLGSrhwjp6lm7GEfauEoSZ1fiOIl
# XdMhSz5SxLVXPyQD8NF6Wy/VI+NwXQ9RRnez+ADhvKwCgl/bwBWzvRvUVUvnOaEP
# 6SNJvBi4RHxF5MHDcnrgcuck379GmcXvwhxX24ON7E1JMKerjt/sW5+v/N2wZuLB
# l4F77dbtS+dJKacTKKanfWeA5opieF+yL4TXV5xcv3coKPHtbcMojyyPQDdPweGF
# RInECUzF1KVDL3SV9274eCBYLBNdYJWaPk8zhNqwiBfenk70lrC8RqBsmNLg1oiM
# CwIDAQABo4IB7TCCAekwEAYJKwYBBAGCNxUBBAMCAQAwHQYDVR0OBBYEFEhuZOVQ
# BdOCqhc3NyK1bajKdQKVMBkGCSsGAQQBgjcUAgQMHgoAUwB1AGIAQwBBMAsGA1Ud
# DwQEAwIBhjAPBgNVHRMBAf8EBTADAQH/MB8GA1UdIwQYMBaAFHItOgIxkEO5FAVO
# 4eqnxzHRI4k0MFoGA1UdHwRTMFEwT6BNoEuGSWh0dHA6Ly9jcmwubWljcm9zb2Z0
# LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01pY1Jvb0NlckF1dDIwMTFfMjAxMV8wM18y
# Mi5jcmwwXgYIKwYBBQUHAQEEUjBQME4GCCsGAQUFBzAChkJodHRwOi8vd3d3Lm1p
# Y3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY1Jvb0NlckF1dDIwMTFfMjAxMV8wM18y
# Mi5jcnQwgZ8GA1UdIASBlzCBlDCBkQYJKwYBBAGCNy4DMIGDMD8GCCsGAQUFBwIB
# FjNodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2RvY3MvcHJpbWFyeWNw
# cy5odG0wQAYIKwYBBQUHAgIwNB4yIB0ATABlAGcAYQBsAF8AcABvAGwAaQBjAHkA
# XwBzAHQAYQB0AGUAbQBlAG4AdAAuIB0wDQYJKoZIhvcNAQELBQADggIBAGfyhqWY
# 4FR5Gi7T2HRnIpsLlhHhY5KZQpZ90nkMkMFlXy4sPvjDctFtg/6+P+gKyju/R6mj
# 82nbY78iNaWXXWWEkH2LRlBV2AySfNIaSxzzPEKLUtCw/WvjPgcuKZvmPRul1LUd
# d5Q54ulkyUQ9eHoj8xN9ppB0g430yyYCRirCihC7pKkFDJvtaPpoLpWgKj8qa1hJ
# Yx8JaW5amJbkg/TAj/NGK978O9C9Ne9uJa7lryft0N3zDq+ZKJeYTQ49C/IIidYf
# wzIY4vDFLc5bnrRJOQrGCsLGra7lstnbFYhRRVg4MnEnGn+x9Cf43iw6IGmYslmJ
# aG5vp7d0w0AFBqYBKig+gj8TTWYLwLNN9eGPfxxvFX1Fp3blQCplo8NdUmKGwx1j
# NpeG39rz+PIWoZon4c2ll9DuXWNB41sHnIc+BncG0QaxdR8UvmFhtfDcxhsEvt9B
# xw4o7t5lL+yX9qFcltgA1qFGvVnzl6UJS0gQmYAf0AApxbGbpT9Fdx41xtKiop96
# eiL6SJUfq/tHI4D1nvi/a7dLl+LrdXga7Oo3mXkYS//WsyNodeav+vyL6wuA6mk7
# r/ww7QRMjt/fdW1jkT3RnVZOT7+AVyKheBEyIXrvQQqxP/uozKRdwaGIm1dxVk5I
# RcBCyZt2WwqASGv9eZ/BvW1taslScxMNelDNMYIWWDCCFlQCAQEwgZUwfjELMAkG
# A1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQx
# HjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEoMCYGA1UEAxMfTWljcm9z
# b2Z0IENvZGUgU2lnbmluZyBQQ0EgMjAxMQITMwAAAVGejY9AcaMOQQAAAAABUTAN
# BglghkgBZQMEAgEFAKCB1jAZBgkqhkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAcBgor
# BgEEAYI3AgELMQ4wDAYKKwYBBAGCNwIBFTAvBgkqhkiG9w0BCQQxIgQgYepWAm3v
# P4sbbGx5aN4ZAK7RlqIIRkHjoE64f5+VqzkwagYKKwYBBAGCNwIBDDFcMFqgPIA6
# AEEAdQB0AG8ARABlAHAAbABvAHkAUgBlAHAAbwByAHQAQQBuAGQAUwB5AG4AYwBE
# AEIALgBwAHMAMaEagBhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcN
# AQEBBQAEggEAWuw7DRKDs6dZ8R/KW+s+9ZivYmEzUMCJpflVhX6ZrbHMjj8/Ww65
# m2IM/8f26aUddDcNhzHASh2sjv2VbQCL9NOax+DexJUqf/cnhgjJZ+Im3ucP5ykz
# EYTtZA25KoQFi3CicYH9IIQytwP7eqEwpqXNP9eCEF6sAsoeEHwuM2whhcNp0MsL
# OiUYYYi3kqjxrFVUfAmGMt9d+ZvCJCxjU76k4QY4Fo+7n1v4FwG9qLt8qdnISWST
# ibsbsH3/AP40EVYZK5fsEayUXnlt0PoCSG4RPsz4GHttuhBYykKyz+TfbAjN2UVe
# cN4yUAMzJOuI3EOm4ZOX6/ShbwZ3aYXWzqGCE7owghO2BgorBgEEAYI3AwMBMYIT
# pjCCE6IGCSqGSIb3DQEHAqCCE5MwghOPAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFY
# BgsqhkiG9w0BCRABBKCCAUcEggFDMIIBPwIBAQYKKwYBBAGEWQoDATAxMA0GCWCG
# SAFlAwQCAQUABCA6YctRmlK1TOuOtVjyNPIKDMt40TZeIyVQ8+hqS+9ABQIGXktG
# IQEoGBMyMDIwMDMyNDA2NTAxMi4zMzNaMAcCAQGAAgH0oIHUpIHRMIHOMQswCQYD
# VQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEe
# MBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSkwJwYDVQQLEyBNaWNyb3Nv
# ZnQgT3BlcmF0aW9ucyBQdWVydG8gUmljbzEmMCQGA1UECxMdVGhhbGVzIFRTUyBF
# U046NTg0Ny1GNzYxLTRGNzAxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1w
# IFNlcnZpY2Wggg8iMIIE9TCCA92gAwIBAgITMwAAAQUHOepZ81W/KgAAAAABBTAN
# BgkqhkiG9w0BAQsFADB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3Rv
# bjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0
# aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMDAeFw0x
# OTA5MDYyMDQxMThaFw0yMDEyMDQyMDQxMThaMIHOMQswCQYDVQQGEwJVUzETMBEG
# A1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWlj
# cm9zb2Z0IENvcnBvcmF0aW9uMSkwJwYDVQQLEyBNaWNyb3NvZnQgT3BlcmF0aW9u
# cyBQdWVydG8gUmljbzEmMCQGA1UECxMdVGhhbGVzIFRTUyBFU046NTg0Ny1GNzYx
# LTRGNzAxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2UwggEi
# MA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDMIpZjVUiLWQGqDFLLaeGfhc9b
# xCwi8HQx+gcaF5Zz2GodhM71oyjal6uqnRM8QHxj49uKFmY/SWEhlV+so3IrmEHV
# LmskeEQaio5PxVgUWRm+sBIJpS9GjwKrGPZ2ub4ST2J9fu5FxbfTmJyB2AL6W7Wc
# SUON8tyuz2/NRAII6YuojdMa6mjVXamL2AS7PBo1GW4Pa4XbNhEQoA4/siS4JGbc
# fAwVMGv87bhKapDqpXLbDq6LbFdLAv7Q7eUHiHS4eccXRNGAnpkdhHOK7s1O9Fqp
# RZYsbx/UpkaoyqiQe/JFTA+1keYZsZgaQqgPNpGYD50SDDyQZ2vtNx7KVgXtAgMB
# AAGjggEbMIIBFzAdBgNVHQ4EFgQUPXQb/rp3KEzK6DYOy3Fn/QIzNyIwHwYDVR0j
# BBgwFoAU1WM6XIoxkPNDe3xGG8UzaFqFbVUwVgYDVR0fBE8wTTBLoEmgR4ZFaHR0
# cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMvTWljVGltU3Rh
# UENBXzIwMTAtMDctMDEuY3JsMFoGCCsGAQUFBwEBBE4wTDBKBggrBgEFBQcwAoY+
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNUaW1TdGFQQ0Ff
# MjAxMC0wNy0wMS5jcnQwDAYDVR0TAQH/BAIwADATBgNVHSUEDDAKBggrBgEFBQcD
# CDANBgkqhkiG9w0BAQsFAAOCAQEAp85hd3trAVfmVAPmcOAqnM47TbWB9S0ySGG/
# eNvIfhgYC0YjCLEZhiiQyOeRTws0lIOWGv7tM5tr70RGzNo1/C7SQadqQ2dT5sUj
# 7Ga6LHO2excdzRvUIwdeOaVuaj4VpiXnhjPBRu2CVNGXPe1d7Zzw7di8xh2D6ooZ
# BjhHLh7yGf2ZFjBjLcDVjrfaLwd4YqefJgg42s8EMUoXzsTpPlS0IBKWeX+RbByc
# OUhXpK9qlvFbBQGp4N+uLEV6haG33oVOtWwrbhu5F0E4UDzsFUaZ8OALyKraR1dI
# o+ZU+zjpn3Na7KUkrT/1UFYdnWYwoUDm9e+DmBhqdhUlxIhRnzCCBnEwggRZoAMC
# AQICCmEJgSoAAAAAAAIwDQYJKoZIhvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xMjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENl
# cnRpZmljYXRlIEF1dGhvcml0eSAyMDEwMB4XDTEwMDcwMTIxMzY1NVoXDTI1MDcw
# MTIxNDY1NVowfDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAO
# BgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEm
# MCQGA1UEAxMdTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwggEiMA0GCSqG
# SIb3DQEBAQUAA4IBDwAwggEKAoIBAQCpHQ28dxGKOiDs/BOX9fp/aZRrdFQQ1aUK
# AIKF++18aEssX8XD5WHCdrc+Zitb8BVTJwQxH0EbGpUdzgkTjnxhMFmxMEQP8WCI
# hFRDDNdNuDgIs0Ldk6zWczBXJoKjRQ3Q6vVHgc2/JGAyWGBG8lhHhjKEHnRhZ5Ff
# gVSxz5NMksHEpl3RYRNuKMYa+YaAu99h/EbBJx0kZxJyGiGKr0tkiVBisV39dx89
# 8Fd1rL2KQk1AUdEPnAY+Z3/1ZsADlkR+79BL/W7lmsqxqPJ6Kgox8NpOBpG2iAg1
# 6HgcsOmZzTznL0S6p/TcZL2kAcEgCZN4zfy8wMlEXV4WnAEFTyJNAgMBAAGjggHm
# MIIB4jAQBgkrBgEEAYI3FQEEAwIBADAdBgNVHQ4EFgQU1WM6XIoxkPNDe3xGG8Uz
# aFqFbVUwGQYJKwYBBAGCNxQCBAweCgBTAHUAYgBDAEEwCwYDVR0PBAQDAgGGMA8G
# A1UdEwEB/wQFMAMBAf8wHwYDVR0jBBgwFoAU1fZWy4/oolxiaNE9lJBb186aGMQw
# VgYDVR0fBE8wTTBLoEmgR4ZFaHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9j
# cmwvcHJvZHVjdHMvTWljUm9vQ2VyQXV0XzIwMTAtMDYtMjMuY3JsMFoGCCsGAQUF
# BwEBBE4wTDBKBggrBgEFBQcwAoY+aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3Br
# aS9jZXJ0cy9NaWNSb29DZXJBdXRfMjAxMC0wNi0yMy5jcnQwgaAGA1UdIAEB/wSB
# lTCBkjCBjwYJKwYBBAGCNy4DMIGBMD0GCCsGAQUFBwIBFjFodHRwOi8vd3d3Lm1p
# Y3Jvc29mdC5jb20vUEtJL2RvY3MvQ1BTL2RlZmF1bHQuaHRtMEAGCCsGAQUFBwIC
# MDQeMiAdAEwAZQBnAGEAbABfAFAAbwBsAGkAYwB5AF8AUwB0AGEAdABlAG0AZQBu
# AHQALiAdMA0GCSqGSIb3DQEBCwUAA4ICAQAH5ohRDeLG4Jg/gXEDPZ2joSFvs+um
# zPUxvs8F4qn++ldtGTCzwsVmyWrf9efweL3HqJ4l4/m87WtUVwgrUYJEEvu5U4zM
# 9GASinbMQEBBm9xcF/9c+V4XNZgkVkt070IQyK+/f8Z/8jd9Wj8c8pl5SpFSAK84
# Dxf1L3mBZdmptWvkx872ynoAb0swRCQiPM/tA6WWj1kpvLb9BOFwnzJKJ/1Vry/+
# tuWOM7tiX5rbV0Dp8c6ZZpCM/2pif93FSguRJuI57BlKcWOdeyFtw5yjojz6f32W
# apB4pm3S4Zz5Hfw42JT0xqUKloakvZ4argRCg7i1gJsiOCC1JeVk7Pf0v35jWSUP
# ei45V3aicaoGig+JFrphpxHLmtgOR5qAxdDNp9DvfYPw4TtxCd9ddJgiCGHasFAe
# b73x4QDf5zEHpJM692VHeOj4qEir995yfmFrb3epgcunCaw5u+zGy9iCtHLNHfS4
# hQEegPsbiSpUObJb2sgNVZl6h3M7COaYLeqN4DMuEin1wC9UJyH3yKxO2ii4sanb
# lrKnQqLJzxlBTeCG+SqaoxFmMNO7dDJL32N79ZmKLxvHIa9Zta7cRDyXUHHXodLF
# VeNp3lfB0d4wwP3M5k37Db9dT+mdHhk4L7zPWAUu7w2gUDXa7wknHNWzfjUeCLra
# NtvTX4/edIhJEqGCA7AwggKYAgEBMIH+oYHUpIHRMIHOMQswCQYDVQQGEwJVUzET
# MBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMV
# TWljcm9zb2Z0IENvcnBvcmF0aW9uMSkwJwYDVQQLEyBNaWNyb3NvZnQgT3BlcmF0
# aW9ucyBQdWVydG8gUmljbzEmMCQGA1UECxMdVGhhbGVzIFRTUyBFU046NTg0Ny1G
# NzYxLTRGNzAxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2Wi
# JQoBATAJBgUrDgMCGgUAAxUA0nmc7MiH2Pr0x33n13Zg4RlV9qqggd4wgdukgdgw
# gdUxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdS
# ZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xKTAnBgNVBAsT
# IE1pY3Jvc29mdCBPcGVyYXRpb25zIFB1ZXJ0byBSaWNvMScwJQYDVQQLEx5uQ2lw
# aGVyIE5UUyBFU046NERFOS0wQzVFLTNFMDkxKzApBgNVBAMTIk1pY3Jvc29mdCBU
# aW1lIFNvdXJjZSBNYXN0ZXIgQ2xvY2swDQYJKoZIhvcNAQEFBQACBQDiJBGpMCIY
# DzIwMjAwMzI0MTI1NzEzWhgPMjAyMDAzMjUxMjU3MTNaMHcwPQYKKwYBBAGEWQoE
# ATEvMC0wCgIFAOIkEakCAQAwCgIBAAICDb4CAf8wBwIBAAICGFkwCgIFAOIlYykC
# AQAwNgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGEWQoDAaAKMAgCAQACAxbjYKEK
# MAgCAQACAwehIDANBgkqhkiG9w0BAQUFAAOCAQEAQMQllxmfvYJFEsKpgIHeFWqu
# dJ59cQPkDUrUUPwzG1GlsxJfcWBFBIfyQsYSAnEdke1/KQkuKQD+mi47t5EW+odJ
# +8vqnjzO0d+WHFaTIf0z20j2sjGNDTYzeKPHlV4PCI9CbErfC7xZntvt7PHynWuV
# iIFuMe9HBef2BcLDp92dg1pzmb2PoDxcs2gme53ztKHoHQpOPn2wb/5xj/z5pvWd
# n1n7TdEyCfWx423cYx/LkqBIcqUbmyN5rOaVJQG7MOeOTVeUVDISqOh45Gw5f9ey
# CpD8gFLQ0L8KkZTOowZATDlSkUr4yfv7+AlfZzopCWjmeZpx4oSkT1iPvKERCDGC
# AvUwggLxAgEBMIGTMHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9u
# MRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRp
# b24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAyMDEwAhMzAAAB
# BQc56lnzVb8qAAAAAAEFMA0GCWCGSAFlAwQCAQUAoIIBMjAaBgkqhkiG9w0BCQMx
# DQYLKoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkEMSIEIJgLtavRcBuNgcFEdLydgJUJ
# tVnUqvxSKruDhwoB5K0mMIHiBgsqhkiG9w0BCRACDDGB0jCBzzCBzDCBsQQU0nmc
# 7MiH2Pr0x33n13Zg4RlV9qowgZgwgYCkfjB8MQswCQYDVQQGEwJVUzETMBEGA1UE
# CBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9z
# b2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQ
# Q0EgMjAxMAITMwAAAQUHOepZ81W/KgAAAAABBTAWBBSdTYJN3k1IHOzgIfYkwU2z
# AuWEVDANBgkqhkiG9w0BAQsFAASCAQA1XDeWKkYm/rvu0usoCriykwgvXVH7aVmI
# MjTclXrJaWkxs4e/N5z3CjSOyMSF0JwPs8Yu5Xr3GrFDmqOBjeX8d7Xl9SdMLxZR
# 0W4GglXeL6iq6r0CozJcTqp+cNuG7StXSm0zJJSk9ASLUy+KNU1mnG1CiXcZbVPs
# 5bsYkk5v423G7iEppgGL7gXWWHMWol6/wKD+TAXEXPaJ9Shwg9WsHbBxSoWoJ4Ug
# KB525f7p9QNPRMHIZ0Ih75wN9VFczY1y9EwtWLOvLxRy34hGKQxMXB7TmIWQ15+G
# VC9Y8h/KfWKoj5ZaALWYmA0Ok9Nv0yp6xhE61yv7l27NGDdg8yWy
# SIG # End signature block
